/**
 * Copyright 2012 the project-owners
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package inti.util;

import java.util.Calendar;
import java.util.Locale;
import java.util.TimeZone;

/**
 * A fast and dedicated DateFormatter, that only applies to the date formats for HTTP related stuff:
 * <br/><br/>
 * Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123<br/>
 * Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036<br/>
 * Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
 */
public class DateFormatter {
    
    /**
     * Weekdays for the RFC 850 format.
     */
    private static final String[] WEEKDAY_850 = new String[] { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday" };
    /**
     * Weekdays for the RFC 1123 format.
     */
    private static final String[] WEEKDAY_1123 = new String[] { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
    /**
     * Month name formats.
     */
    private static final String[] MONTH_NAME = new String[] { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
            "Sep", "Oct", "Nov", "Dec" };
    /**
     * Thread-safe calendar instances.
     */
    private static final ThreadLocal<Calendar> CALENDARS = new ThreadLocal<Calendar>() {
        
        @Override
        protected Calendar initialValue() {
            return Calendar.getInstance(TimeZone.getTimeZone("GMT"), Locale.ENGLISH);
        }
        
    };
    /**
     * Index correction number for fast array-element access.
     */
    private static final int MONTH_CORRECTION = 'A' + 'a' + 'a';
    /**
     * Array that addresses months by the first three letters.
     */
    private static final int[] MONTH_INDEX = new int[] { -1, -1, -1, -1, -1, -1, -1, -1, -1, Calendar.DECEMBER,
            Calendar.FEBRUARY, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, Calendar.JANUARY, -1, -1, -1,
            Calendar.AUGUST, -1, -1, Calendar.MARCH, -1, -1, Calendar.APRIL, -1, -1, Calendar.OCTOBER, Calendar.MAY,
            Calendar.SEPTEMBER, -1, -1, Calendar.JULY, -1, Calendar.JUNE, -1, -1, -1, -1, -1, Calendar.NOVEMBER };
    
    /**
     * Enlisting of valid date formats.
     */
    public enum Format {
        RFC850, RFC1123, ANSI_C
    }
    
    /**
     * Calls {@link #parseDate(String, Calendar)}.
     * 
     * @param date The date string to parse.
     * @return The date as milliseconds.
     */
    public long parseDate(String date) {
        return parseDate(date, CALENDARS.get());
    }
    
    /**
     * Parses the date string using their respective first unique sign to determine the format used.
     * 
     * @param date The date string to parse.
     * @param cal The calendar instance to use.
     * @return The date as milliseconds.
     */
    public long parseDate(String date, Calendar cal) {
        int index = date.indexOf(','), day, month, year, minute, hour, second;
        char[] chars = date.toCharArray();
        
        day = month = year = minute = hour = second = 0;
        
        if (index == 3) {
            // RFC822 / RFC1123
            day = (chars[5] - '0') * 10 + chars[6] - '0';
            month = MONTH_INDEX[chars[8] + chars[9] + chars[10] - MONTH_CORRECTION];
            year = (chars[12] - '0') * 1000 + (chars[13] - '0') * 100 + (chars[14] - '0') * 10 + chars[15] - '0';
            hour = (chars[17] - '0') * 10 + chars[18] - '0';
            minute = (chars[20] - '0') * 10 + chars[21] - '0';
            second = (chars[23] - '0') * 10 + chars[24] - '0';
        } else if (index == -1) {
            // ASCI_C
            day = (chars[8] >= '0' && chars[8] <= '9' ? chars[8] - '0' : 0) * 10 + chars[9] - '0';
            month = MONTH_INDEX[chars[4] + chars[5] + chars[6] - MONTH_CORRECTION];
            year = (chars[20] - '0') * 1000 + (chars[21] - '0') * 100 + (chars[22] - '0') * 10 + chars[23] - '0';
            hour = (chars[11] - '0') * 10 + chars[12] - '0';
            minute = (chars[14] - '0') * 10 + chars[15] - '0';
            second = (chars[17] - '0') * 10 + chars[18] - '0';
        } else {
            // RFC850
            day = (chars[index + 2] - '0') * 10 + chars[index + 3] - '0';
            month = MONTH_INDEX[chars[index + 5] + chars[index + 6] + chars[index + 7] - MONTH_CORRECTION];
            year = (chars[index + 9] - '0') * 10 + chars[index + 10] - '0';
            year += year > 50 ? 1900 : 2000;
            hour = (chars[index + 12] - '0') * 10 + chars[index + 13] - '0';
            minute = (chars[index + 15] - '0') * 10 + chars[index + 16] - '0';
            second = (chars[index + 18] - '0') * 10 + chars[index + 19] - '0';
        }
        
        cal.setTimeInMillis(0);
        cal.set(Calendar.YEAR, year);
        cal.set(Calendar.MONTH, month);
        cal.set(Calendar.DAY_OF_MONTH, day);
        cal.set(Calendar.HOUR_OF_DAY, hour - 1);
        cal.set(Calendar.MINUTE, minute);
        cal.set(Calendar.SECOND, second);
        
        return cal.getTimeInMillis();
    }
    
    /**
     * Formats the specified date in {@link Format#RFC1123} - creates a new string builder and a new string.
     * 
     * @param date The date in milliseconds.
     * @return The formated date.
     */
    public String formatDate(long date) {
        StringBuilder builder = new StringBuilder();
        format1123(date, builder, CALENDARS.get());
        return builder.toString();
    }
    
    /**
     * Formats the specified date in {@link Format#RFC1123} - uses the string builder and creates a new string.
     * 
     * @param date The date in milliseconds.
     * @param builder The string builder used as buffer.
     * @return The formated date.
     */
    public void formatDate(long date, StringBuilder builder) {
        format1123(date, builder, CALENDARS.get());
    }
    
    /**
     * Formats the specified date in given format - uses the string builder and creates a new string.
     * 
     * @param date The date in milliseconds.
     * @param builder The string builder used as buffer.
     * @param format The format to use.
     * @return The formated date.
     */
    public void formatDate(long date, Format format, StringBuilder builder) {
        formatDate(date, format, builder, CALENDARS.get());
    }
    
    /**
     * Formats the specified date in given format - uses the string builder and creates a new string.
     * <br/><br/>
     * http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3.1
     * <br/><br/>
     * Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123<br/>
     * Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036<br/>
     * Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format
     * 
     * @param date The date in milliseconds.
     * @param builder The string builder used as buffer.
     * @param format The format to use.
     * @param cal The calendar instance.
     * @return The formated date.
     */
    public void formatDate(long date, Format format, StringBuilder builder, Calendar cal) {
        switch (format) {
            case ANSI_C:
                asci_c(date, builder, cal);
                break;
            case RFC850:
                format850(date, builder, cal);
                break;
            default:
            case RFC1123:
                format1123(date, builder, cal);
                break;
        }
    }
    
    /**
     * Formats the given date in asci_c format.
     */
    protected void asci_c(long date, StringBuilder builder, Calendar cal) {
        int tmp;
        
        cal.setTimeInMillis(date);
        
        builder.append(WEEKDAY_1123[cal.get(Calendar.DAY_OF_WEEK) - 1]);
        builder.append(' ');
        builder.append(MONTH_NAME[cal.get(Calendar.MONTH)]);
        builder.append(' ');
        tmp = cal.get(Calendar.DAY_OF_MONTH);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        
        builder.append(' ');
        tmp = cal.get(Calendar.HOUR_OF_DAY);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.MINUTE);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.SECOND);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(' ');
        builder.append(cal.get(Calendar.YEAR));
    }
    
    /**
     * Formats the given date in rfc1123 format.
     */
    protected void format1123(long date, StringBuilder builder, Calendar cal) {
        int tmp;
        
        cal.setTimeInMillis(date);
        
        builder.append(WEEKDAY_1123[cal.get(Calendar.DAY_OF_WEEK) - 1]);
        builder.append(", ");
        tmp = cal.get(Calendar.DAY_OF_MONTH);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(' ');
        builder.append(MONTH_NAME[cal.get(Calendar.MONTH)]);
        builder.append(' ');
        builder.append(cal.get(Calendar.YEAR));
        
        builder.append(' ');
        tmp = cal.get(Calendar.HOUR_OF_DAY);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.MINUTE);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.SECOND);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(" GMT");
    }
    
    /**
     * Formats the given date in rfc850 format.
     */
    protected void format850(long date, StringBuilder builder, Calendar cal) {
        int tmp;
        
        cal.setTimeInMillis(date);
        
        builder.append(WEEKDAY_850[cal.get(Calendar.DAY_OF_WEEK) - 1]);
        builder.append(", ");
        tmp = cal.get(Calendar.DAY_OF_MONTH);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append('-');
        builder.append(MONTH_NAME[cal.get(Calendar.MONTH)]);
        builder.append('-');
        tmp = cal.get(Calendar.YEAR) % 100;
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        
        builder.append(' ');
        tmp = cal.get(Calendar.HOUR_OF_DAY);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.MINUTE);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(':');
        tmp = cal.get(Calendar.SECOND);
        if (tmp < 10) {
            builder.append('0');
        }
        builder.append(tmp);
        builder.append(" GMT");
    }
    
}
