/**
 * Copyright 2012 the project-owners
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */
package inti.util;

import java.io.IOException;
import java.io.InputStream;

/**
 * InputStream which notifies on reads and closes.
 */
public class NotifyingInputStream extends InputStream {
    
    /**
     * Notifier interface for input streams.
     */
    public interface ReadNotifier {
        
        void read(int b);
        
        void read(byte[] data, int offset, int length);
        
        void close();
        
    }
    
    InputStream source;
    ReadNotifier notifier;
    
    @Override
    public int read() throws IOException {
        int data = source.read();
        notifier.read(data);
        return data;
    }
    
    @Override
    public int read(byte[] b, int off, int len) throws IOException {
        int dataRead = source.read(b, off, len);
        notifier.read(b, off, dataRead);
        return dataRead;
    }
    
    @Override
    public int available() throws IOException {
        return source.available();
    }
    
    @Override
    public void close() throws IOException {
        notifier.close();
        source.close();
    }
    
    @Override
    public void mark(int readlimit) {
        source.mark(readlimit);
    }
    
    @Override
    public boolean markSupported() {
        return source.markSupported();
    }
    
    @Override
    public void reset() throws IOException {
        source.reset();
    }
    
    @Override
    public long skip(long n) throws IOException {
        return source.skip(n);
    }
    
    public InputStream getSource() {
        return source;
    }
    
    public void setSource(InputStream source) {
        this.source = source;
    }
    
    public ReadNotifier getNotifier() {
        return notifier;
    }
    
    public void setNotifier(ReadNotifier notifier) {
        this.notifier = notifier;
    }
    
}
